using System.Security;
using System.Text;
using System.Text.RegularExpressions;
using System.Web;

namespace UpdateWiki.Helpers
{
    class MarkdownConverter
    {
        private const string OrderedListHtmlTagName = "ol";
        private const string UnorderedListHtmlTagName = "ul";

        class BlockInfo
        {
            public BlockElement BlockElement { get; set; }
            public int? Indention { get; set; }

            public BlockInfo(BlockElement blockElement)
            {
                BlockElement = blockElement;
                Indention = null;
            }

            public BlockInfo(BlockElement blockElement, int indention)
            {
                BlockElement = blockElement;
                Indention = indention;
            }
        }

        enum BlockElement
        {
            UnorderedList,
            OrderedList,
            CodeBlock,
            PlantUmlDiagram,
            Table
        }

        public static string ConvertMarkdownFileToHtml(string dirName, string markdownFile, bool embedImages)
        {
            StringBuilder htmlContent = new();
            string[] markdownFileLines = File.ReadAllLines(markdownFile);
            int currentLine = 0, totalLineCount = markdownFileLines.Length;
            bool isMetadataBlock = true, isParagrah = false;
            Stack<BlockInfo> blocks = new();

            while (currentLine < totalLineCount)
            {
                string line = markdownFileLines[currentLine].TrimEnd();
                if (isMetadataBlock) isMetadataBlock = IsMetadataBlock(line);

                if (!isMetadataBlock)
                {
                    string closingTags = string.Empty;
                    line = ParseInlineElements(line);
                    bool isLineParsed = TryParseLineElements(line.TrimStart(), out string newLine);

                    if (isLineParsed)
                    {
                        line = newLine;
                    }

                    bool isBlockParsed = TryParseBlockElements(dirName, line, blocks, embedImages, out string newLine2);

                    if (isBlockParsed)
                    {
                        line = newLine2;
                    }

                    if (!isParagrah && !string.IsNullOrWhiteSpace(line) && !isLineParsed && !isBlockParsed)
                    {
                        line = "<p>" + line;
                        isParagrah = true;
                    }

                    if (isParagrah && (string.IsNullOrWhiteSpace(line) || isLineParsed || isBlockParsed || currentLine == totalLineCount - 1))
                    {
                        if (currentLine == totalLineCount - 1)
                        {
                            line += "</p>";
                        }
                        else
                        {
                            line = "</p>" + line;
                        }
                        
                        isParagrah = false;
                    }

                    if (!isBlockParsed || currentLine == totalLineCount - 1)
                    {
                        CloseBlockElements(blocks, out closingTags);
                    }

                    if (currentLine < totalLineCount - 1)
                    {
                        htmlContent.Append(closingTags + line);
                    }
                    else
                    {
                        htmlContent.Append(line + closingTags);
                    }
                }

                currentLine++;
            }

            return htmlContent.ToString();
        }

        private static bool IsMetadataBlock(string line)
        {
            return string.IsNullOrWhiteSpace(line) || line.TrimStart().StartsWith("<!-- ");
        }

        private static string ParseInlineElements(string line)
        {
            string ouputLine = SecurityElement.Escape(line);
            ouputLine = ParseBold(ouputLine);
            ouputLine = ParseItalic(ouputLine);
            ouputLine = ParseCode(ouputLine);
            ouputLine = ParseImage(ouputLine);
            ouputLine = ParseLink(ouputLine);

            return ouputLine;
        }

        private static string ParseBold(string line)
        {
            return ParseSymbol(line, @"\*\*", "b");
        }

        private static string ParseItalic(string line)
        {
            string newLine = ParseSymbol(line, @"\*", "i");

            return ParseSymbol(newLine, "_", "i");
        }

        private static string ParseCode(string line)
        {
            string? outputLine = null;

            if (!line.StartsWith("```"))
            {
                outputLine = ParseSymbol(line, "`", "code");
            }

            return outputLine ?? line;
        }

        private static string ParseImage(string line)
        {
            string outputLine = new(line);
            Match? match;

            do
            {
                match = Regex.Match(outputLine, @"!\[([^\]]+)\]\(([^\)]+)\)");

                if (match.Success)
                {
                    Regex regex = new(@"!\[[^\]]+\]\([^\)]+\)");
                    outputLine = regex.Replace(outputLine, string.Format(@"<img src=""{0}"" alt=""{1}"" />", match.Groups[2].Value, match.Groups[1].Value), 1);
                }
            } while (match.Success);

            return outputLine;
        }

        private static string ParseLink(string line)
        {
            string newLine = ParseMarkdownLink(line);

            return ParseRawLink(newLine);
        }

        private static string ParseMarkdownLink(string line)
        {
            string outputLine = new(line);
            Match? match;

            do
            {
                match = Regex.Match(outputLine, @"\[([^\]]+)\]\(([^\)]+)\)");

                if (match.Success)
                {
                    Regex regex = new(@"\[[^\]]+\]\([^\)]+\)");
                    outputLine = regex.Replace(outputLine, string.Format(@"<a href=""{0}"">{1}</a>", match.Groups[2].Value, match.Groups[1].Value), 1);
                }
            } while (match.Success);

            return outputLine;
        }

        private static string ParseRawLink(string line)
        {
            string outputLine = new(line);
            Match? match;

            do
            {
                match = Regex.Match(outputLine, "([^\">])(https?://(www\\.)?[^\\s]+\\.[^\\s]+)([^\"<])");

                if (match.Success)
                {
                    Regex regex = new("([^\">])(https?://(www\\.)?[^\\s]+\\.[^\\s]+)([^\"<])");
                    outputLine = regex.Replace(outputLine, string.Format(@"{0}<a href=""{1}"">{2}</a>{3}", match.Groups[1].Value, match.Groups[2].Value, match.Groups[2].Value, match.Groups[match.Groups.Count - 1].Value), 1);
                }
            } while (match.Success);

            return outputLine;
        }

        private static string ParseSymbol(string line, string markdownSymbolPattern, string htmlTagName)
        {
            string outputLine = new(line);
            string markdownSymbol = Regex.Unescape(markdownSymbolPattern);
            int boldSymbolCount = Regex.Matches(outputLine, markdownSymbolPattern).Count;

            while (boldSymbolCount > 1)
            {
                outputLine = ReplaceFirst(ReplaceFirst(outputLine, markdownSymbol, $"<{htmlTagName}>"), markdownSymbol, $"</{htmlTagName}>");
                boldSymbolCount = Regex.Matches(outputLine, markdownSymbolPattern).Count;
            }

            return outputLine;
        }

        private static string ReplaceFirst(string text, string search, string replace)
        {
            int pos = text.IndexOf(search);

            if (pos < 0)
            {
                return text;
            }

            return string.Concat(text.AsSpan(0, pos), replace, text.AsSpan(pos + search.Length));
        }

        private static bool TryParseLineElements(string line, out string parsedLine)
        {
            bool parsed = TryParseHeader(line, out parsedLine);
            if (!parsed) parsed = TryParseBlockquote(line, out parsedLine);

            return parsed;
        }

        private static bool TryParseHeader(string line, out string parsedLine)
        {
            bool isParsed = false;
            StringBuilder outputLine = new();

            if (line.StartsWith("### "))
            {
                outputLine.Append(line.Replace("### ", "<h3>"));
                outputLine.Append("</h3>");
                isParsed = true;
            }
            else if (line.StartsWith("## "))
            {
                outputLine.Append(line.Replace("## ", "<h2>"));
                outputLine.Append("</h2>");
                isParsed = true;
            }
            else if (line.StartsWith("# "))
            {
                outputLine.Append(line.Replace("# ", "<h1>"));
                outputLine.Append("</h1>");
                isParsed = true;
            }

            parsedLine = outputLine.ToString();

            return isParsed;
        }

        private static bool TryParseBlockquote(string line, out string parsedLine)
        {
            bool isParsed = false;
            parsedLine = string.Empty;

            if (line.StartsWith("&gt; ") && line.Length > 5)
            {
                parsedLine = string.Concat("<blockquote>", line.AsSpan(5), "</blockquote>");
                isParsed = true;
            }

            return isParsed;
        }

        private static bool TryParseBlockElements(string dirName, string line, Stack<BlockInfo> blocks, bool embedImages, out string parsedLine)
        {
            bool isParsed = TryParseOrderedList(line, blocks, out parsedLine);
            if (!isParsed) isParsed = TryParseUnorderedList(line, blocks, out parsedLine);
            if (!isParsed) isParsed = TryParsePlantUmlDiagram(dirName, line.TrimStart(), blocks, embedImages, out parsedLine);
            if (!isParsed) isParsed = TryParseCodeBlock(line, blocks, out parsedLine);
            if (!isParsed) isParsed = TryParseTable(line.Trim(), blocks, out parsedLine);

            return isParsed;
        }

        private static void CloseBlockElements(Stack<BlockInfo> blocks, out string closingTags)
        {
            StringBuilder closingTagsBuilder = new();

            while (blocks.Count > 0)
            {
                BlockInfo block = blocks.Pop();

                switch (block.BlockElement)
                {
                    case BlockElement.UnorderedList:
                        closingTagsBuilder.Append("</li></" + UnorderedListHtmlTagName + ">");
                        break;
                    case BlockElement.OrderedList:
                        closingTagsBuilder.Append("</li></" + OrderedListHtmlTagName + ">");
                        break;
                    case BlockElement.Table:
                        closingTagsBuilder.Append("</table>");
                        break;
                }
            }

            closingTags = closingTagsBuilder.ToString();
        }

        private static bool TryParseOrderedList(string line, Stack<BlockInfo> blocks, out string parsedLine)
        {
            return TryParseList(line, blocks, "^( *)[0-9]+. ", BlockElement.OrderedList, out parsedLine);
        }

        private static bool TryParseUnorderedList(string line, Stack<BlockInfo> blocks, out string parsedLine)
        {
            return TryParseList(line, blocks, "^( *)- ", BlockElement.UnorderedList, out parsedLine);
        }

        private static bool TryParseList(string line, Stack<BlockInfo> blocks, string markdownPattern, BlockElement listBlockElement, out string parsedLine)
        {
            bool isParsed = false;
            parsedLine = string.Empty;
            blocks.TryPeek(out BlockInfo? blockInfo);
            Match match = Regex.Match(line, markdownPattern);

            if (match.Success)
            {
                parsedLine = Regex.Replace(line, markdownPattern, string.Empty);
                int indention = match.Groups[1].Length == 0 ? 0 : (match.Groups[1].Length / 4);

                if (blockInfo is null || (indention > blockInfo.Indention)) // First list item
                {
                    string htmlTagName = listBlockElement.Equals(BlockElement.OrderedList) ? OrderedListHtmlTagName : UnorderedListHtmlTagName;
                    parsedLine = "<" + htmlTagName + "><li>" + parsedLine;
                    blockInfo = new BlockInfo(listBlockElement, indention);
                    blocks.Push(blockInfo);
                }
                else if (indention < blockInfo.Indention) // First item after nested list
                {
                    int removedIndentionLevels = blockInfo.Indention.Value - indention;
                    StringBuilder blockClosings = new();

                    for(int i = 0; i < removedIndentionLevels; i++)
                    {
                        blockInfo = blocks.Pop();
                        string htmlTagName = blockInfo.BlockElement.Equals(BlockElement.OrderedList) ? OrderedListHtmlTagName : UnorderedListHtmlTagName;
                        blockClosings.Append("</" + htmlTagName + "></li>");
                    }

                    parsedLine = "</li>" + blockClosings.ToString() + "<li>" + parsedLine;
                }
                else
                {
                    parsedLine = "</li><li>" + parsedLine;
                }
                
                isParsed = true;
            }

            return isParsed;
        }

        private static bool TryParsePlantUmlDiagram(string dirName, string line, Stack<BlockInfo> blocks, bool embedImages, out string parsedLine)
        {
            bool isParsed = false;
            parsedLine = string.Empty;
            blocks.TryPeek(out BlockInfo? blockInfo);

            if (line.StartsWith("```plantuml") && blockInfo is null)
            {
                blockInfo = new BlockInfo(BlockElement.PlantUmlDiagram);
                blocks.Push(blockInfo);
                isParsed = true;
            }
            else if (line.StartsWith("```") && blockInfo is not null && blockInfo.BlockElement.Equals(BlockElement.PlantUmlDiagram))
            {
                blocks.Pop();
                isParsed = true;
            }
            else if (blockInfo is not null && blockInfo.BlockElement.Equals(BlockElement.PlantUmlDiagram))
            {
                string? diagramName = null;
                string textToSearch = "@startuml ";

                if (line.StartsWith(textToSearch))
                {
                    diagramName = line[(line.IndexOf(textToSearch) + textToSearch.Length)..];
                }

                if (!string.IsNullOrEmpty(diagramName))
                {
                    if (embedImages)
                    {
                        byte[] imageArray = File.ReadAllBytes(Path.Combine(dirName, diagramName + ".svg"));
                        string base64ImageRepresentation = Convert.ToBase64String(imageArray);
                        parsedLine = $@"<img src=""data:image/svg+xml;base64, {base64ImageRepresentation}"" alt=""{diagramName}"" />";
                    }
                    else
                    {
                        parsedLine = @"<ac:image><ri:attachment ri:filename=""" + diagramName + @".svg"" /></ac:image>";
                    }
                }

                isParsed = true;
            }

            return isParsed;
        }

        private static bool TryParseCodeBlock(string line, Stack<BlockInfo> blocks, out string parsedLine)
        {
            bool isParsed = false;
            parsedLine = string.Empty;
            blocks.TryPeek(out BlockInfo? blockInfo);

            if (line.TrimStart().StartsWith("`") && blockInfo is null)
            {
                parsedLine = "<code><pre>";
                blockInfo = new BlockInfo(BlockElement.CodeBlock);
                blocks.Push(blockInfo);
                isParsed = true;
            }
            else if (line.TrimStart().StartsWith("`") && blockInfo is not null && blockInfo.BlockElement.Equals(BlockElement.CodeBlock))
            {
                parsedLine = "</pre></code>";
                blocks.Pop();
                isParsed = true;
            }
            else if (blockInfo is not null && blockInfo.BlockElement.Equals(BlockElement.CodeBlock))
            {
                parsedLine = HttpUtility.HtmlEncode(line) + "\n";
                isParsed = true; // Inside code block
            }

            return isParsed;
        }

        private static bool TryParseTable(string line, Stack<BlockInfo> blocks, out string parsedLine)
        {
            bool isParsed = false;
            parsedLine = string.Empty;
            blocks.TryPeek(out BlockInfo? blockInfo);

            if (line.StartsWith("|") && blockInfo is null)
            {
                parsedLine = "<table><tr>" + line.Replace(" | ", "</th><th>").Replace("| ", "<th>").Replace(" |", "</th>") + "</tr>";
                blockInfo = new BlockInfo(BlockElement.Table);
                blocks.Push(blockInfo);
                isParsed = true;
            }
            else if (line.Replace(" ", string.Empty).StartsWith("|---") && blockInfo is not null && blockInfo.BlockElement.Equals(BlockElement.Table))
            {
                parsedLine = string.Empty;
                isParsed = true;
            }
            else if (line.StartsWith("|") && blockInfo is not null && blockInfo.BlockElement.Equals(BlockElement.Table))
            {
                parsedLine = "<tr><td>" + line[1..^1].Replace("|", "</td><td>") + "</td></tr>";
                isParsed = true;
            }

            return isParsed;
        }
    }
}